@@ -37,7 +37,6 @@ use std::ptr::NonNull;
 use num_enum::IntoPrimitive;
 use num_enum::TryFromPrimitive;
 
-use crate::libbpf_sys;
 use crate::util::create_bpf_entity_checked;
 use crate::util::create_bpf_entity_checked_opt;
 use crate::util::parse_ret_i32;
@@ -194,10 +193,9 @@ impl Btf<'static> {
 }
 
 impl<'btf> Btf<'btf> {
-    pub(crate) fn from_bpf_object(obj: &'btf libbpf_sys::bpf_object) -> Result<Self> {
-        Self::from_bpf_object_raw(obj).and_then(|opt| {
-            opt.ok_or_else(|| Error::with_io_error(io::ErrorKind::NotFound, "btf not found"))
-        })
+    /// Create a new `Btf` instance from the given [`libbpf_sys::bpf_object`].
+    pub fn from_bpf_object(obj: &'btf libbpf_sys::bpf_object) -> Result<Option<Self>> {
+        Self::from_bpf_object_raw(obj)
     }
 
     fn from_bpf_object_raw(obj: *const libbpf_sys::bpf_object) -> Result<Option<Self>> {
