@@ -258,15 +258,19 @@ pub enum MemberAttr {
 
 impl MemberAttr {
     #[inline]
-    fn normal(offset: u32) -> Self {
-        Self::Normal { offset }
-    }
-
-    #[inline]
-    fn bif_field(offset: u32) -> Self {
-        Self::BitField {
-            size: (offset >> 24) as u8,
-            offset: offset & 0x00_ff_ff_ff,
+    fn new(kflag: bool, offset: u32) -> Self {
+        if kflag {
+            let size = (offset >> 24) as u8;
+            if size != 0 {
+                Self::BitField {
+                    size,
+                    offset: offset & 0x00_ff_ff_ff,
+                }
+            } else {
+                Self::Normal { offset }
+            }
+        } else {
+            Self::Normal { offset }
         }
     }
 }
@@ -458,11 +462,7 @@ gen_collection_concrete_type! {
     |btf, member, kflag| StructMember {
         name: btf.name_at(member.name_off),
         ty: member.type_.into(),
-        attr: if kflag {
-            MemberAttr::bif_field(member.offset)
-        } else {
-            MemberAttr::normal(member.offset)
-        },
+        attr: MemberAttr::new(kflag, member.offset),
     }
 }
 
@@ -486,17 +486,13 @@ gen_collection_concrete_type! {
     |btf, member, kflag| UnionMember {
         name: btf.name_at(member.name_off),
         ty: member.type_.into(),
-        attr: if kflag {
-            MemberAttr::bif_field(member.offset)
-        } else {
-            MemberAttr::normal(member.offset)
-        },
+        attr: MemberAttr::new(kflag, member.offset),
     }
 }
 
 /// A Composite type, which can be one of a [`Struct`] or a [`Union`].
 ///
-/// Sometimes it's not usefull to distinguish them, in that case, one can use this
+/// Sometimes it's not useful to distinguish them, in that case, one can use this
 /// type to inspect any of them.
 #[derive(Debug)]
 pub struct Composite<'btf> {
@@ -583,11 +579,7 @@ gen_collection_members_concrete_type! {
     |btf, member, kflag| CompositeMember {
         name: btf.name_at(member.name_off),
         ty: member.type_.into(),
-        attr: if kflag {
-            MemberAttr::bif_field(member.offset)
-        } else {
-            MemberAttr::normal(member.offset)
-        },
+        attr: MemberAttr::new(kflag, member.offset),
     }
 }
 
@@ -996,6 +988,8 @@ macro_rules! __btf_type_match {
 
 #[cfg(test)]
 mod test {
+    use super::*;
+
     // creates a dummy btftype, not it's not safe to use this type, but it is safe to match on it,
     // which is all we need for these tests.
     macro_rules! dummy_type {
@@ -1018,8 +1012,6 @@ mod test {
         "int"
     }
 
-    use super::BtfType;
-    use crate::btf_type_match;
     #[test]
     fn full_switch_case() {
         dummy_type!(ty);
