@@ -123,7 +123,9 @@ pub use crate::print::get_print;
 pub use crate::print::set_print;
 pub use crate::print::PrintCallback;
 pub use crate::print::PrintLevel;
+pub use crate::program::Input as ProgramInput;
 pub use crate::program::OpenProgram;
+pub use crate::program::Output as ProgramOutput;
 pub use crate::program::Program;
 pub use crate::program::ProgramAttachType;
 pub use crate::program::ProgramType;
