@@ -1,10 +1,10 @@
 use core::ffi::c_void;
-use std::convert::TryFrom;
 use std::ffi::CStr;
 use std::ffi::CString;
 use std::ffi::OsStr;
 use std::fmt::Debug;
 use std::fs::remove_file;
+use std::io;
 use std::mem;
 use std::ops::Deref;
 use std::os::unix::ffi::OsStrExt;
@@ -17,13 +17,12 @@ use std::os::unix::io::RawFd;
 use std::path::Path;
 use std::ptr;
 use std::ptr::NonNull;
+use std::slice;
 use std::slice::from_raw_parts;
 
 use bitflags::bitflags;
 use libbpf_sys::bpf_map_info;
 use libbpf_sys::bpf_obj_get_info_by_fd;
-use nix::errno;
-use nix::unistd;
 use num_enum::IntoPrimitive;
 use num_enum::TryFromPrimitive;
 use strum_macros::Display;
@@ -74,6 +73,36 @@ impl OpenMap {
         }
     }
 
+    fn initial_value_raw(&self) -> (*mut u8, usize) {
+        let mut size = 0u64;
+        let ptr = unsafe {
+            libbpf_sys::bpf_map__initial_value(self.ptr.as_ptr(), &mut size as *mut _ as _)
+        };
+        (ptr.cast(), size as _)
+    }
+
+    /// Retrieve the initial value of the map.
+    pub fn initial_value(&self) -> Option<&[u8]> {
+        let (ptr, size) = self.initial_value_raw();
+        if ptr.is_null() {
+            None
+        } else {
+            let data = unsafe { slice::from_raw_parts(ptr.cast::<u8>(), size) };
+            Some(data)
+        }
+    }
+
+    /// Retrieve the initial value of the map.
+    pub fn initial_value_mut(&mut self) -> Option<&mut [u8]> {
+        let (ptr, size) = self.initial_value_raw();
+        if ptr.is_null() {
+            None
+        } else {
+            let data = unsafe { slice::from_raw_parts_mut(ptr.cast::<u8>(), size) };
+            Some(data)
+        }
+    }
+
     pub fn set_map_ifindex(&mut self, idx: u32) {
         unsafe { libbpf_sys::bpf_map__set_ifindex(self.ptr.as_ptr(), idx) };
     }
@@ -145,8 +174,8 @@ impl OpenMap {
     }
 
     /// Reuse an fd for a BPF map
-    pub fn reuse_fd(&self, fd: i32) -> Result<()> {
-        let ret = unsafe { libbpf_sys::bpf_map__reuse_fd(self.ptr.as_ptr(), fd) };
+    pub fn reuse_fd(&self, fd: BorrowedFd<'_>) -> Result<()> {
+        let ret = unsafe { libbpf_sys::bpf_map__reuse_fd(self.ptr.as_ptr(), fd.as_raw_fd()) };
         util::parse_ret(ret)
     }
 
@@ -156,14 +185,12 @@ impl OpenMap {
 
         let fd = unsafe { libbpf_sys::bpf_obj_get(cstring.as_ptr()) };
         if fd < 0 {
-            return Err(Error::from_raw_os_error(errno::errno()));
+            return Err(Error::from(io::Error::last_os_error()));
         }
-        let reuse_result = self.reuse_fd(fd);
 
-        // Always close `fd` regardless of if `reuse_fd` succeeded or failed
-        //
-        // Ignore errors b/c can't really recover from failure
-        let _ = unistd::close(fd);
+        let fd = unsafe { OwnedFd::from_raw_fd(fd) };
+
+        let reuse_result = self.reuse_fd(fd.as_fd());
 
         reuse_result
     }
@@ -237,9 +264,7 @@ impl Map {
 
         // Get the map fd
         let fd = unsafe { libbpf_sys::bpf_map__fd(ptr.as_ptr()) };
-        if fd < 0 {
-            return Err(Error::from_raw_os_error(-fd));
-        }
+        let fd = util::parse_ret_i32(fd)?;
 
         let ty = MapType::try_from(unsafe { libbpf_sys::bpf_map__type(ptr.as_ptr()) })
             .unwrap_or(MapType::Unknown);
@@ -568,11 +593,11 @@ impl MapHandle {
             }
             Ok(Some(out))
         } else {
-            let errno = errno::errno();
-            if errno::Errno::from_i32(errno) == errno::Errno::ENOENT {
+            let err = io::Error::last_os_error();
+            if err.kind() == io::ErrorKind::NotFound {
                 Ok(None)
             } else {
-                Err(Error::from_raw_os_error(errno))
+                Err(Error::from(err))
             }
         }
     }
@@ -682,10 +707,13 @@ impl MapHandle {
             )));
         };
 
+        #[allow(clippy::needless_update)]
         let opts = libbpf_sys::bpf_map_batch_opts {
             sz: mem::size_of::<libbpf_sys::bpf_map_batch_opts>() as _,
             elem_flags: elem_flags.bits(),
             flags: flags.bits(),
+            // bpf_map_batch_opts might have padding fields on some platform
+            ..Default::default()
         };
 
         let mut count = count;
@@ -731,11 +759,11 @@ impl MapHandle {
             }
             Ok(Some(out))
         } else {
-            let errno = errno::errno();
-            if errno::Errno::from_i32(errno) == errno::Errno::ENOENT {
+            let err = io::Error::last_os_error();
+            if err.kind() == io::ErrorKind::NotFound {
                 Ok(None)
             } else {
-                Err(Error::from_raw_os_error(errno))
+                Err(Error::from(err))
             }
         }
     }
@@ -795,10 +823,13 @@ impl MapHandle {
             )));
         }
 
+        #[allow(clippy::needless_update)]
         let opts = libbpf_sys::bpf_map_batch_opts {
             sz: mem::size_of::<libbpf_sys::bpf_map_batch_opts>() as _,
             elem_flags: elem_flags.bits(),
             flags: flags.bits(),
+            // bpf_map_batch_opts might have padding fields on some platform
+            ..Default::default()
         };
 
         let mut count = count;
