@@ -1,7 +1,8 @@
-use core::ffi::c_void;
-use std::convert::TryFrom;
+use std::ffi::c_void;
 use std::ffi::CStr;
+use std::mem;
 use std::mem::size_of;
+use std::mem::size_of_val;
 use std::os::unix::io::AsFd;
 use std::os::unix::io::AsRawFd;
 use std::os::unix::io::BorrowedFd;
@@ -16,7 +17,6 @@ use libbpf_sys::bpf_func_id;
 use num_enum::TryFromPrimitive;
 use strum_macros::Display;
 
-use crate::libbpf_sys;
 use crate::util;
 use crate::AsRawLibbpf;
 use crate::Error;
@@ -60,9 +60,12 @@ impl From<UsdtOpts> for libbpf_sys::bpf_usdt_opts {
             cookie,
             _non_exhaustive,
         } = opts;
+        #[allow(clippy::needless_update)]
         libbpf_sys::bpf_usdt_opts {
             sz: size_of::<Self>() as _,
             usdt_cookie: cookie,
+            // bpf_usdt_opts might have padding fields on some platform
+            ..Default::default()
         }
     }
 }
@@ -83,9 +86,12 @@ impl From<TracepointOpts> for libbpf_sys::bpf_tracepoint_opts {
             _non_exhaustive,
         } = opts;
 
+        #[allow(clippy::needless_update)]
         libbpf_sys::bpf_tracepoint_opts {
             sz: size_of::<Self>() as _,
             bpf_cookie: cookie,
+            // bpf_tracepoint_opts might have padding fields on some platform
+            ..Default::default()
         }
     }
 }
@@ -231,7 +237,7 @@ impl OpenProgram {
     }
 }
 
-impl AsRawLibbpf for Program {
+impl AsRawLibbpf for OpenProgram {
     type LibbpfType = libbpf_sys::bpf_program;
 
     /// Retrieve the underlying [`libbpf_sys::bpf_program`].
@@ -366,6 +372,46 @@ pub enum ProgramAttachType {
     Unknown = u32::MAX,
 }
 
+/// The input a program accepts.
+///
+/// This type is mostly used in conjunction with the [`Program::test_run`]
+/// facility.
+#[derive(Debug, Default)]
+pub struct Input<'dat> {
+    /// The input context to provide.
+    pub context_in: Option<&'dat [u8]>,
+    /// The output context buffer provided to the program.
+    pub context_out: Option<&'dat mut [u8]>,
+    /// Additional data to provide to the program.
+    pub data_in: Option<&'dat [u8]>,
+    /// The output data buffer provided to the program.
+    pub data_out: Option<&'dat mut [u8]>,
+    /// The 'cpu' value passed to the kernel.
+    pub cpu: u32,
+    /// The 'flags' value passed to the kernel.
+    pub flags: u32,
+    /// The struct is non-exhaustive and open to extension.
+    #[doc(hidden)]
+    pub _non_exhaustive: (),
+}
+
+/// The output a program produces.
+///
+/// This type is mostly used in conjunction with the [`Program::test_run`]
+/// facility.
+#[derive(Debug)]
+pub struct Output<'dat> {
+    /// The value returned by the program.
+    pub return_value: u32,
+    /// The output context filled by the program/kernel.
+    pub context: Option<&'dat mut [u8]>,
+    /// Output data filled by the program.
+    pub data: Option<&'dat mut [u8]>,
+    /// The struct is non-exhaustive and open to extension.
+    #[doc(hidden)]
+    pub _non_exhaustive: (),
+}
+
 /// Represents a loaded [`Program`].
 ///
 /// This struct is not safe to clone because the underlying libbpf resource cannot currently
@@ -614,6 +660,29 @@ impl Program {
         })
     }
 
+    /// Attach this program to the specified syscall
+    pub fn attach_ksyscall<T: AsRef<str>>(
+        &mut self,
+        retprobe: bool,
+        syscall_name: T,
+    ) -> Result<Link> {
+        let opts = libbpf_sys::bpf_ksyscall_opts {
+            sz: size_of::<libbpf_sys::bpf_ksyscall_opts>() as _,
+            retprobe,
+            ..Default::default()
+        };
+
+        let syscall_name = util::str_to_cstring(syscall_name.as_ref())?;
+        let syscall_name_ptr = syscall_name.as_ptr();
+        util::create_bpf_entity_checked(|| unsafe {
+            libbpf_sys::bpf_program__attach_ksyscall(self.ptr.as_ptr(), syscall_name_ptr, &opts)
+        })
+        .map(|ptr| unsafe {
+            // SAFETY: the pointer came from libbpf and has been checked for errors
+            Link::new(ptr)
+        })
+    }
+
     fn attach_tracepoint_impl(
         &mut self,
         tp_category: &str,
@@ -845,6 +914,67 @@ impl Program {
         })
     }
 
+    /// Test run the program with the given input data.
+    ///
+    /// This function uses the
+    /// [BPF_PROG_RUN](https://www.kernel.org/doc/html/latest/bpf/bpf_prog_run.html)
+    /// facility.
+    pub fn test_run<'dat>(&mut self, input: Input<'dat>) -> Result<Output<'dat>> {
+        pub(crate) unsafe fn slice_from_array<'t, T>(
+            items: *mut T,
+            num_items: usize,
+        ) -> Option<&'t mut [T]> {
+            if items.is_null() {
+                None
+            } else {
+                Some(unsafe { slice::from_raw_parts_mut(items, num_items) })
+            }
+        }
+
+        let Input {
+            context_in,
+            mut context_out,
+            data_in,
+            mut data_out,
+            cpu,
+            flags,
+            _non_exhaustive: (),
+        } = input;
+
+        let mut opts = unsafe { mem::zeroed::<libbpf_sys::bpf_test_run_opts>() };
+        opts.sz = size_of_val(&opts) as _;
+        opts.ctx_in = context_in
+            .map(|data| data.as_ptr().cast())
+            .unwrap_or_else(ptr::null);
+        opts.ctx_size_in = context_in.map(|data| data.len() as _).unwrap_or(0);
+        opts.ctx_out = context_out
+            .as_mut()
+            .map(|data| data.as_mut_ptr().cast())
+            .unwrap_or_else(ptr::null_mut);
+        opts.ctx_size_out = context_out.map(|data| data.len() as _).unwrap_or(0);
+        opts.data_in = data_in
+            .map(|data| data.as_ptr().cast())
+            .unwrap_or_else(ptr::null);
+        opts.data_size_in = data_in.map(|data| data.len() as _).unwrap_or(0);
+        opts.data_out = data_out
+            .as_mut()
+            .map(|data| data.as_mut_ptr().cast())
+            .unwrap_or_else(ptr::null_mut);
+        opts.data_size_out = data_out.map(|data| data.len() as _).unwrap_or(0);
+        opts.cpu = cpu;
+        opts.flags = flags;
+
+        let rc = unsafe { libbpf_sys::bpf_prog_test_run_opts(self.as_fd().as_raw_fd(), &mut opts) };
+        let () = util::parse_ret(rc)?;
+        let output = Output {
+            return_value: opts.retval,
+            context: unsafe { slice_from_array(opts.ctx_out.cast(), opts.ctx_size_out as _) },
+            data: unsafe { slice_from_array(opts.data_out.cast(), opts.data_size_out as _) },
+            _non_exhaustive: (),
+        };
+        Ok(output)
+    }
+
     /// Returns the number of instructions that form the program.
     ///
     /// Please see note in [`OpenProgram::insn_cnt`].
@@ -862,3 +992,12 @@ impl Program {
         unsafe { slice::from_raw_parts(ptr, count) }
     }
 }
+
+impl AsRawLibbpf for Program {
+    type LibbpfType = libbpf_sys::bpf_program;
+
+    /// Retrieve the underlying [`libbpf_sys::bpf_program`].
+    fn as_libbpf_object(&self) -> NonNull<Self::LibbpfType> {
+        self.ptr
+    }
+}
