@@ -10,18 +10,19 @@
 //! }
 //! ```
 
-use std::convert::TryFrom;
 use std::ffi::c_void;
 use std::ffi::CString;
+use std::io;
 use std::mem::size_of_val;
+use std::os::fd::AsFd;
+use std::os::fd::AsRawFd;
+use std::os::fd::BorrowedFd;
+use std::os::fd::FromRawFd;
+use std::os::fd::OwnedFd;
 use std::os::raw::c_char;
 use std::ptr;
 use std::time::Duration;
 
-use nix::errno;
-use nix::unistd::close;
-
-use crate::libbpf_sys;
 use crate::util;
 use crate::MapType;
 use crate::ProgramAttachType;
@@ -40,7 +41,7 @@ macro_rules! gen_info_impl {
 
         impl $name {
             // Returns Some(next_valid_fd), None on none left
-            fn next_valid_fd(&mut self) -> Option<i32> {
+            fn next_valid_fd(&mut self) -> Option<OwnedFd> {
                 loop {
                     if unsafe { $next_id(self.cur_id, &mut self.cur_id) } != 0 {
                         return None;
@@ -48,14 +49,15 @@ macro_rules! gen_info_impl {
 
                     let fd = unsafe { $fd_by_id(self.cur_id) };
                     if fd < 0 {
-                        if errno::errno() == errno::Errno::ENOENT as i32 {
+                        let err = io::Error::last_os_error();
+                        if err.kind() == io::ErrorKind::NotFound {
                             continue;
                         }
 
                         return None;
                     }
 
-                    return Some(fd);
+                    return Some(unsafe { OwnedFd::from_raw_fd(fd)});
                 }
             }
         }
@@ -64,10 +66,7 @@ macro_rules! gen_info_impl {
             type Item = $info_ty;
 
             fn next(&mut self) -> Option<Self::Item> {
-                let fd = match self.next_valid_fd() {
-                    Some(fd) => fd,
-                    None => return None,
-                };
+                let fd = self.next_valid_fd()?;
 
                 // We need to use std::mem::zeroed() instead of just using
                 // ::default() because padding bytes need to be zero as well.
@@ -78,14 +77,13 @@ macro_rules! gen_info_impl {
                 let item_ptr: *mut $uapi_info_ty = &mut item;
                 let mut len = size_of_val(&item) as u32;
 
-                let ret = unsafe { libbpf_sys::bpf_obj_get_info_by_fd(fd, item_ptr as *mut c_void, &mut len) };
+                let ret = unsafe { libbpf_sys::bpf_obj_get_info_by_fd(fd.as_raw_fd(), item_ptr as *mut c_void, &mut len) };
                 let parsed_uapi = if ret != 0 {
                     None
                 } else {
-                    <$info_ty>::from_uapi(fd, item)
+                    <$info_ty>::from_uapi(fd.as_fd(), item)
                 };
 
-                let _ = close(fd);
                 parsed_uapi
             }
         }
@@ -155,6 +153,8 @@ pub struct ProgramInfo {
     pub prog_tags: Vec<Tag>,
     pub run_time_ns: u64,
     pub run_cnt: u64,
+    /// Skipped BPF executions due to recursion or concurrent execution prevention.
+    pub recursion_misses: u64,
 }
 
 /// An iterator for the information of loaded bpf programs
@@ -269,7 +269,7 @@ impl ProgInfoQueryOptions {
 }
 
 impl ProgramInfo {
-    fn load_from_fd(fd: i32, opts: &ProgInfoQueryOptions) -> Result<Self> {
+    fn load_from_fd(fd: BorrowedFd<'_>, opts: &ProgInfoQueryOptions) -> Result<Self> {
         let mut item = libbpf_sys::bpf_prog_info::default();
 
         let mut xlated_prog_insns: Vec<u8> = Vec::new();
@@ -285,8 +285,9 @@ impl ProgramInfo {
         let item_ptr: *mut libbpf_sys::bpf_prog_info = &mut item;
         let mut len = size_of_val(&item) as u32;
 
-        let ret =
-            unsafe { libbpf_sys::bpf_obj_get_info_by_fd(fd, item_ptr as *mut c_void, &mut len) };
+        let ret = unsafe {
+            libbpf_sys::bpf_obj_get_info_by_fd(fd.as_raw_fd(), item_ptr as *mut c_void, &mut len)
+        };
         util::parse_ret(ret)?;
 
         // SANITY: `libbpf` should guarantee NUL termination.
@@ -365,8 +366,9 @@ impl ProgramInfo {
             item.nr_jited_ksyms = 0;
         }
 
-        let ret =
-            unsafe { libbpf_sys::bpf_obj_get_info_by_fd(fd, item_ptr as *mut c_void, &mut len) };
+        let ret = unsafe {
+            libbpf_sys::bpf_obj_get_info_by_fd(fd.as_raw_fd(), item_ptr as *mut c_void, &mut len)
+        };
         util::parse_ret(ret)?;
 
         return Ok(ProgramInfo {
@@ -395,14 +397,13 @@ impl ProgramInfo {
             prog_tags,
             run_time_ns: item.run_time_ns,
             run_cnt: item.run_cnt,
+            recursion_misses: item.recursion_misses,
         });
     }
 }
 
-impl Iterator for ProgInfoIter {
-    type Item = ProgramInfo;
-
-    fn next(&mut self) -> Option<Self::Item> {
+impl ProgInfoIter {
+    fn next_valid_fd(&mut self) -> Option<OwnedFd> {
         loop {
             if unsafe { libbpf_sys::bpf_prog_get_next_id(self.cur_id, &mut self.cur_id) } != 0 {
                 return None;
@@ -410,20 +411,30 @@ impl Iterator for ProgInfoIter {
 
             let fd = unsafe { libbpf_sys::bpf_prog_get_fd_by_id(self.cur_id) };
             if fd < 0 {
-                if errno::errno() == errno::Errno::ENOENT as i32 {
+                let err = io::Error::last_os_error();
+                if err.kind() == io::ErrorKind::NotFound {
                     continue;
                 }
                 return None;
             }
 
-            let prog = ProgramInfo::load_from_fd(fd, &self.opts);
-            let _ = close(fd);
+            return Some(unsafe { OwnedFd::from_raw_fd(fd) });
+        }
+    }
+}
 
-            match prog {
-                Ok(p) => return Some(p),
-                // TODO: We should consider bubbling up errors properly.
-                Err(_err) => (),
-            }
+impl Iterator for ProgInfoIter {
+    type Item = ProgramInfo;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        let fd = self.next_valid_fd()?;
+
+        let prog = ProgramInfo::load_from_fd(fd.as_fd(), &self.opts);
+
+        match prog {
+            Ok(p) => Some(p),
+            // TODO: We should consider bubbling up errors properly.
+            Err(_err) => None,
         }
     }
 }
@@ -450,7 +461,7 @@ pub struct MapInfo {
 }
 
 impl MapInfo {
-    fn from_uapi(_fd: i32, s: libbpf_sys::bpf_map_info) -> Option<Self> {
+    fn from_uapi(_fd: BorrowedFd<'_>, s: libbpf_sys::bpf_map_info) -> Option<Self> {
         // SANITY: `libbpf` should guarantee NUL termination.
         let name = util::c_char_slice_to_cstr(&s.name).unwrap();
         let ty = match MapType::try_from(s.type_) {
@@ -498,7 +509,7 @@ pub struct BtfInfo {
 }
 
 impl BtfInfo {
-    fn load_from_fd(fd: i32) -> Result<Self> {
+    fn load_from_fd(fd: BorrowedFd<'_>) -> Result<Self> {
         let mut item = libbpf_sys::bpf_btf_info::default();
         let mut btf: Vec<u8> = Vec::new();
         let mut name: Vec<u8> = Vec::new();
@@ -506,8 +517,9 @@ impl BtfInfo {
         let item_ptr: *mut libbpf_sys::bpf_btf_info = &mut item;
         let mut len = size_of_val(&item) as u32;
 
-        let ret =
-            unsafe { libbpf_sys::bpf_obj_get_info_by_fd(fd, item_ptr as *mut c_void, &mut len) };
+        let ret = unsafe {
+            libbpf_sys::bpf_obj_get_info_by_fd(fd.as_raw_fd(), item_ptr as *mut c_void, &mut len)
+        };
         util::parse_ret(ret)?;
 
         // The API gives you the ascii string length while expecting
@@ -519,8 +531,9 @@ impl BtfInfo {
         btf.resize(item.btf_size as usize, 0u8);
         item.btf = btf.as_mut_ptr() as *mut c_void as u64;
 
-        let ret =
-            unsafe { libbpf_sys::bpf_obj_get_info_by_fd(fd, item_ptr as *mut c_void, &mut len) };
+        let ret = unsafe {
+            libbpf_sys::bpf_obj_get_info_by_fd(fd.as_raw_fd(), item_ptr as *mut c_void, &mut len)
+        };
         util::parse_ret(ret)?;
 
         Ok(BtfInfo {
@@ -541,10 +554,9 @@ pub struct BtfInfoIter {
     cur_id: u32,
 }
 
-impl Iterator for BtfInfoIter {
-    type Item = BtfInfo;
-
-    fn next(&mut self) -> Option<Self::Item> {
+impl BtfInfoIter {
+    // Returns Some(next_valid_fd), None on none left
+    fn next_valid_fd(&mut self) -> Option<OwnedFd> {
         loop {
             if unsafe { libbpf_sys::bpf_btf_get_next_id(self.cur_id, &mut self.cur_id) } != 0 {
                 return None;
@@ -552,20 +564,30 @@ impl Iterator for BtfInfoIter {
 
             let fd = unsafe { libbpf_sys::bpf_btf_get_fd_by_id(self.cur_id) };
             if fd < 0 {
-                if errno::errno() == errno::Errno::ENOENT as i32 {
+                let err = io::Error::last_os_error();
+                if err.kind() == io::ErrorKind::NotFound {
                     continue;
                 }
                 return None;
             }
 
-            let info = BtfInfo::load_from_fd(fd);
-            let _ = close(fd);
+            return Some(unsafe { OwnedFd::from_raw_fd(fd) });
+        }
+    }
+}
 
-            match info {
-                Ok(i) => return Some(i),
-                // TODO: We should consider bubbling up errors properly.
-                Err(_err) => (),
-            }
+impl Iterator for BtfInfoIter {
+    type Item = BtfInfo;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        let fd = self.next_valid_fd()?;
+
+        let info = BtfInfo::load_from_fd(fd.as_fd());
+
+        match info {
+            Ok(i) => Some(i),
+            // TODO: We should consider bubbling up errors properly.
+            Err(_err) => None,
         }
     }
 }
@@ -623,7 +645,7 @@ pub struct LinkInfo {
 }
 
 impl LinkInfo {
-    fn from_uapi(fd: i32, mut s: libbpf_sys::bpf_link_info) -> Option<Self> {
+    fn from_uapi(fd: BorrowedFd<'_>, mut s: libbpf_sys::bpf_link_info) -> Option<Self> {
         let type_info = match s.type_ {
             libbpf_sys::BPF_LINK_TYPE_RAW_TRACEPOINT => {
                 let mut buf = [0; 256];
@@ -633,7 +655,11 @@ impl LinkInfo {
                 let mut len = size_of_val(&s) as u32;
 
                 let ret = unsafe {
-                    libbpf_sys::bpf_obj_get_info_by_fd(fd, item_ptr as *mut c_void, &mut len)
+                    libbpf_sys::bpf_obj_get_info_by_fd(
+                        fd.as_raw_fd(),
+                        item_ptr as *mut c_void,
+                        &mut len,
+                    )
                 };
                 if ret != 0 {
                     return None;
