@@ -1,21 +1,23 @@
 #![allow(clippy::let_unit_value)]
+#![warn(clippy::absolute_paths)]
 
 use std::collections::HashSet;
+use std::env::current_exe;
+use std::ffi::c_int;
 use std::ffi::c_void;
 use std::fs;
 use std::hint;
+use std::io;
 use std::io::Read;
 use std::mem::size_of;
-use std::os::fd::AsRawFd;
 use std::os::unix::io::AsFd;
 use std::path::Path;
 use std::path::PathBuf;
+use std::ptr::addr_of;
 use std::slice;
 use std::sync::mpsc::channel;
 use std::time::Duration;
 
-use nix::errno;
-use nix::unistd::close;
 use plain::Plain;
 use probe::probe;
 use scopeguard::defer;
@@ -35,6 +37,7 @@ use libbpf_rs::Object;
 use libbpf_rs::ObjectBuilder;
 use libbpf_rs::OpenObject;
 use libbpf_rs::Program;
+use libbpf_rs::ProgramInput;
 use libbpf_rs::ProgramType;
 use libbpf_rs::TracepointOpts;
 use libbpf_rs::UprobeOpts;
@@ -78,7 +81,7 @@ pub fn bump_rlimit_mlock() {
         ret,
         0,
         "Setting RLIMIT_MEMLOCK failed with errno: {}",
-        errno::errno()
+        io::Error::last_os_error()
     );
 }
 
@@ -336,6 +339,7 @@ fn test_sudo_object_map_delete_batch() {
 /// Test whether `MapInfo` works properly
 #[test]
 pub fn test_sudo_map_info() {
+    #[allow(clippy::needless_update)]
     let opts = libbpf_sys::bpf_map_create_opts {
         sz: size_of::<libbpf_sys::bpf_map_create_opts>() as libbpf_sys::size_t,
         map_flags: libbpf_sys::BPF_ANY,
@@ -347,6 +351,8 @@ pub fn test_sudo_map_info() {
         map_extra: 0,
         numa_node: 0,
         map_ifindex: 0,
+        // bpf_map_create_opts might have padding fields on some platform
+        ..Default::default()
     };
 
     let map = MapHandle::create(MapType::Hash, Some("simple_map"), 8, 64, 1024, &opts).unwrap();
@@ -844,6 +850,147 @@ fn test_sudo_object_reuse_pined_map() {
     assert!(!Path::new(path).exists());
 }
 
+#[test]
+fn test_sudo_object_ringbuf_raw() {
+    bump_rlimit_mlock();
+
+    let mut obj = get_test_object("ringbuf.bpf.o");
+    let prog = obj
+        .prog_mut("handle__sys_enter_getpid")
+        .expect("failed to find program");
+    let _link = prog.attach().expect("failed to attach prog");
+
+    static mut V1: i32 = 0;
+    static mut V2: i32 = 0;
+
+    fn callback1(data: &[u8]) -> i32 {
+        let mut value: i32 = 0;
+        plain::copy_from_bytes(&mut value, data).expect("Wrong size");
+
+        unsafe {
+            V1 = value;
+        }
+
+        0
+    }
+
+    fn callback2(data: &[u8]) -> i32 {
+        let mut value: i32 = 0;
+        plain::copy_from_bytes(&mut value, data).expect("Wrong size");
+
+        unsafe {
+            V2 = value;
+        }
+
+        0
+    }
+
+    // Test trying to build without adding any ringbufs
+    // Can't use expect_err here since RingBuffer does not implement Debug
+    let builder = libbpf_rs::RingBufferBuilder::new();
+    assert!(
+        builder.build().is_err(),
+        "Should not be able to build without adding at least one ringbuf"
+    );
+
+    // Test building with multiple map objects
+    let mut builder = libbpf_rs::RingBufferBuilder::new();
+
+    // Add a first map and callback
+    let map1 = obj.map("ringbuf1").expect("Failed to get ringbuf1 map");
+
+    builder.add(map1, callback1).expect("Failed to add ringbuf");
+
+    // Add a second map and callback
+    let map2 = obj.map("ringbuf2").expect("Failed to get ringbuf2 map");
+
+    builder.add(map2, callback2).expect("Failed to add ringbuf");
+
+    let mgr = builder.build().expect("Failed to build");
+
+    // Call getpid to ensure the BPF program runs
+    unsafe { libc::getpid() };
+
+    // Test raw primitives
+    let ret = mgr.consume_raw();
+
+    // We can't check for exact return values, since other tasks in the system may call getpid(),
+    // triggering the BPF program
+    assert!(ret >= 2);
+
+    unsafe { assert_eq!(V1, 1) };
+    unsafe { assert_eq!(V2, 2) };
+
+    // Consume from a (potentially) empty ring buffer
+    let ret = mgr.consume_raw();
+    assert!(ret >= 0);
+
+    // Consume from a (potentially) empty ring buffer using poll()
+    let ret = mgr.poll_raw(Duration::from_millis(100));
+    assert!(ret >= 0);
+}
+
+#[test]
+fn test_sudo_object_ringbuf_err_callback() {
+    bump_rlimit_mlock();
+
+    let mut obj = get_test_object("ringbuf.bpf.o");
+    let prog = obj
+        .prog_mut("handle__sys_enter_getpid")
+        .expect("failed to find program");
+    let _link = prog.attach().expect("failed to attach prog");
+
+    // Immediately trigger an error that should be reported back to the consume_raw() or poll_raw()
+    fn callback1(_data: &[u8]) -> i32 {
+        -libc::ENOENT
+    }
+
+    // Immediately trigger an error that should be reported back to the consume_raw() or poll_raw()
+    fn callback2(_data: &[u8]) -> i32 {
+        -libc::EPERM
+    }
+
+    // Test trying to build without adding any ringbufs
+    // Can't use expect_err here since RingBuffer does not implement Debug
+    let builder = libbpf_rs::RingBufferBuilder::new();
+    assert!(
+        builder.build().is_err(),
+        "Should not be able to build without adding at least one ringbuf"
+    );
+
+    // Test building with multiple map objects
+    let mut builder = libbpf_rs::RingBufferBuilder::new();
+
+    // Add a first map and callback
+    let map1 = obj.map("ringbuf1").expect("Failed to get ringbuf1 map");
+
+    builder.add(map1, callback1).expect("Failed to add ringbuf");
+
+    // Add a second map and callback
+    let map2 = obj.map("ringbuf2").expect("Failed to get ringbuf2 map");
+
+    builder.add(map2, callback2).expect("Failed to add ringbuf");
+
+    let mgr = builder.build().expect("Failed to build");
+
+    // Call getpid to ensure the BPF program runs
+    unsafe { libc::getpid() };
+
+    // Test raw primitives
+    let ret = mgr.consume_raw();
+
+    // The error originated from the first callback executed should be reported here, either
+    // from callback1() or callback2()
+    assert!(ret == -libc::ENOENT || ret == -libc::EPERM);
+
+    unsafe { libc::getpid() };
+
+    // The same behavior should happen with poll_raw()
+    let ret = mgr.poll_raw(Duration::from_millis(100));
+
+    assert!(ret == -libc::ENOENT || ret == -libc::EPERM);
+}
+
 #[test]
 fn test_sudo_object_ringbuf() {
     bump_rlimit_mlock();
@@ -1180,6 +1327,7 @@ fn test_sudo_object_map_create_and_pin() {
 fn test_sudo_object_map_create_without_name() {
     bump_rlimit_mlock();
 
+    #[allow(clippy::needless_update)]
     let opts = libbpf_sys::bpf_map_create_opts {
         sz: size_of::<libbpf_sys::bpf_map_create_opts>() as libbpf_sys::size_t,
         map_flags: libbpf_sys::BPF_F_NO_PREALLOC,
@@ -1191,6 +1339,8 @@ fn test_sudo_object_map_create_without_name() {
         map_extra: 0,
         numa_node: 0,
         map_ifindex: 0,
+        // bpf_map_create_opts might have padding fields on some platform
+        ..Default::default()
     };
 
     let map = MapHandle::create(MapType::Hash, Option::<&str>::None, 4, 8, 8, &opts)
@@ -1245,7 +1395,7 @@ fn test_sudo_object_usdt() {
         .prog_mut("handle__usdt")
         .expect("Failed to find program");
 
-    let path = std::env::current_exe().expect("Failed to find executable name");
+    let path = current_exe().expect("Failed to find executable name");
     let _link = prog
         .attach_usdt(
             unsafe { libc::getpid() },
@@ -1275,7 +1425,7 @@ fn test_sudo_object_usdt_cookie() {
         .prog_mut("handle__usdt_with_cookie")
         .expect("Failed to find program");
 
-    let path = std::env::current_exe().expect("Failed to find executable name");
+    let path = current_exe().expect("Failed to find executable name");
     let _link = prog
         .attach_usdt_with_opts(
             unsafe { libc::getpid() },
@@ -1434,7 +1584,7 @@ fn test_sudo_object_uprobe_with_opts() {
         .expect("Failed to find program");
 
     let pid = unsafe { libc::getpid() };
-    let path = std::env::current_exe().expect("Failed to find executable name");
+    let path = current_exe().expect("Failed to find executable name");
     let func_offset = 0;
     let opts = UprobeOpts {
         func_name: "uprobe_target".to_string(),
@@ -1466,7 +1616,7 @@ fn test_sudo_object_uprobe_with_cookie() {
         .expect("Failed to find program");
 
     let pid = unsafe { libc::getpid() };
-    let path = std::env::current_exe().expect("Failed to find executable name");
+    let path = current_exe().expect("Failed to find executable name");
     let func_offset = 0;
     let opts = UprobeOpts {
         func_name: "uprobe_target".to_string(),
@@ -1633,8 +1783,7 @@ fn test_sudo_program_get_fd_and_id() {
 
     let prog_fd = prog.as_fd();
     let prog_id = Program::get_id_by_fd(prog_fd).expect("failed to get program id by fd");
-    let owned_prog_fd = Program::get_fd_by_id(prog_id).expect("failed to get program fd by id");
-    close(owned_prog_fd.as_raw_fd()).expect("failed to close owned program fd");
+    let _owned_prog_fd = Program::get_fd_by_id(prog_id).expect("failed to get program fd by id");
 }
 
 /// Check that autocreate disabled maps don't prevent object loading
@@ -1652,3 +1801,87 @@ fn test_sudo_map_autocreate_disable() {
 
     open_obj.load().expect("failed to load object");
 }
+
+/// Check that we can resize a map.
+#[test]
+fn test_sudo_map_resize() {
+    bump_rlimit_mlock();
+
+    let mut open_obj = open_test_object("map_auto_pin.bpf.o");
+
+    let resizable = open_obj
+        .map_mut(".data.resizable_data")
+        .expect("map wasn't found");
+
+    let len = resizable.initial_value().unwrap().len();
+    assert_eq!(len, size_of::<u64>());
+
+    let () = resizable
+        .set_value_size(len as u32 * 2)
+        .expect("failed to set value size");
+    let new_len = resizable.initial_value().unwrap().len();
+    assert_eq!(new_len, len * 2);
+}
+
+/// Check that we are able to attach using ksyscall
+#[test]
+fn test_sudo_attach_ksyscall() {
+    bump_rlimit_mlock();
+
+    let mut obj = get_test_object("ksyscall.bpf.o");
+    let prog = obj
+        .prog_mut("handle__ksyscall")
+        .expect("Failed to find program");
+
+    let _link = prog
+        .attach_ksyscall(false, "kill")
+        .expect("Failed to attach prog");
+
+    let map = obj.map("ringbuf").expect("Failed to get ringbuf map");
+    let action = || {
+        // Send `SIGCHLD`, which is ignored by default, to our process.
+        let ret = unsafe { libc::kill(libc::getpid(), libc::SIGCHLD) };
+        if ret < 0 {
+            panic!("kill failed: {}", io::Error::last_os_error());
+        }
+    };
+    let result = with_ringbuffer(map, action);
+
+    assert_eq!(result, 1);
+}
+
+/// Check that we can invoke a program directly.
+#[test]
+fn test_sudo_run_prog_success() {
+    bump_rlimit_mlock();
+
+    let mut obj = get_test_object("run_prog.bpf.o");
+    let prog = obj.prog_mut("test_1").expect("Failed to find program");
+
+    #[repr(C)]
+    struct bpf_dummy_ops_state {
+        val: c_int,
+    }
+
+    let value = 42;
+    let state = bpf_dummy_ops_state { val: value };
+    let args = [addr_of!(state) as u64];
+    let input = ProgramInput {
+        context_in: Some(unsafe { plain::as_bytes(&args) }),
+        ..Default::default()
+    };
+    let output = prog.test_run(input).unwrap();
+    assert_eq!(output.return_value, value as _);
+}
+
+/// Check that we fail program invocation when providing insufficient arguments.
+#[test]
+fn test_sudo_run_prog_fail() {
+    bump_rlimit_mlock();
+
+    let mut obj = get_test_object("run_prog.bpf.o");
+    let prog = obj.prog_mut("test_2").expect("Failed to find program");
+
+    let input = ProgramInput::default();
+    let _err = prog.test_run(input).unwrap_err();
+}
